# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .x2imagenet import EasyData2ImageNet, JingLing2ImageNet
from .x2seg import JingLing2Seg, LabelMe2Seg, EasyData2Seg
from .x2voc import LabelMe2VOC, EasyData2VOC
from .x2coco import LabelMe2COCO, EasyData2COCO, JingLing2COCO

__all__ = [
    'EasyData2ImageNet', 'JingLing2ImageNet', 'JingLing2Seg', 'LabelMe2Seg',
    'EasyData2Seg', 'LabelMe2VOC', 'EasyData2VOC', 'LabelMe2COCO',
    'EasyData2COCO', 'JingLing2COCO'
]
